﻿Public Class Clamp
    Inherits FunctionBase

    Private fMin As Double = 0
    Private fMax As Double = 1
    Private fFilteredSignal As Double = 0
    Private fScale As Double = 1
    Private fAdd As Double = 0

    Public Overrides Function Evaluate(x As Double, y As Double) As Double
        If Double.IsNaN(fFilteredSignal) Then
            fFilteredSignal = fAdd
        End If
        Dim fResult As Double = fFilteredSignal + fAdd
        fResult *= fScale
        If fResult < fMin Then
            Return fMin
        ElseIf fResult > fMax Then
            Return fMax
        Else
            Return fResult
        End If
    End Function

    Public Overrides ReadOnly Property Name As String
        Get
            Return "Clamp"
        End Get
    End Property

    Public Overrides ReadOnly Property Type As FunctionType
        Get
            Return FunctionType.Function
        End Get
    End Property

    Public Overrides ReadOnly Property ParameterInfos As DataTable
        Get
            Dim tbl As New DataTable
            tbl.Columns.Add("Index", GetType(Int32))
            tbl.Columns.Add("Name", GetType(String))
            tbl.Columns.Add("Step", GetType(Double))
            tbl.Columns.Add("Default", GetType(Double))
            tbl.Columns.Add("Current", GetType(Double))

            tbl.Rows.Add(0, "Add", 0.1, 0, fAdd)
            tbl.Rows.Add(1, "Scale", 0.1, 1, fScale)
            tbl.Rows.Add(2, "Minimum", 0.1, 0, fMin)
            tbl.Rows.Add(3, "Maximum", 0.1, 1, fMax)
            tbl.Rows.Add(4, "Signal", 0.1, 0, fFilteredSignal)
            Return tbl
        End Get
    End Property

    Public Overrides ReadOnly Property Help As String
        Get
            Return "Liegt das Signal unter Minimum wird Minimum ausgegeben, liegt das Signal über Maximum, wird Maximum ausgegeben. Sonst Signal."
        End Get
    End Property


    Public Overrides Sub SetParameterValue(Index As Integer, Value As Double)
        Select Case Index
            Case 0
                fAdd = Value
            Case 1
                fScale = Value
            Case 2
                fMin = Value
            Case 3
                fMax = Value
            Case 4
                fFilteredSignal = Value
        End Select
    End Sub

    Public Overrides Sub AddParameterValue(Index As Integer, Value As Double)
        Select Case Index
            Case 0
                fAdd += Value
            Case 1
                fScale += Value
            Case 2
                fMin += Value
            Case 3
                fMax += Value
            Case 4
                fFilteredSignal += Value
        End Select
    End Sub
End Class
