﻿Public Class Crossfade
    Inherits FunctionBase

    Private fAdd As Double = 0
    Private fIntensity As Double = 1

    Private fSwitch As Double = 0
    Private fSignal1 As Double = 0
    Private fSignal2 As Double = 0



    Public Overrides Function Evaluate(ByVal x As Double, ByVal y As Double) As Double
        Dim fTemp As Double = fSwitch
        If fTemp < 0 Then
            fTemp = 0
        ElseIf fTemp > 1 Then
            fTemp = 1
        End If
        Dim fResult As Double = fSignal1 * (1 - fTemp) + fSignal2 * fTemp
        
        Return (fResult + fAdd) * fIntensity
    End Function

    Public Overrides ReadOnly Property Help As String
        Get
            Return "Das Ergebnis ist Signal 1 * (1 - Switch) + Signal 2 * Switch. Wobei Switch auf 0-1 beschränkt wird."
        End Get
    End Property

    Public Overrides ReadOnly Property Name As String
        Get
            Return "Crossfade"
        End Get
    End Property

    Public Overrides ReadOnly Property ParameterInfos As System.Data.DataTable
        Get
            Dim tbl As New DataTable
            tbl.Columns.Add("Index", GetType(Int32))
            tbl.Columns.Add("Name", GetType(String))
            tbl.Columns.Add("Step", GetType(Double))
            tbl.Columns.Add("Default", GetType(Double))
            tbl.Columns.Add("Current", GetType(Double))
            tbl.Rows.Add(0, "Add", 0.1, 0, fAdd)
            tbl.Rows.Add(1, "Intensity", 0.1, 1, fIntensity)

            tbl.Rows.Add(2, "Switch", 0.1, 0, fSwitch)
            tbl.Rows.Add(3, "Signal 1", 0.1, 0, fSignal1)
            tbl.Rows.Add(4, "Signal 2", 0.1, 0, fSignal2)
            Return tbl
        End Get
    End Property

    Public Overloads Overrides Sub SetParameterValue(ByVal Index As Integer, ByVal Value As Double)
        Select Case Index
            Case 0
                fAdd = Value
            Case 1
                fIntensity = Value
            Case 2
                fSwitch = Value
            Case 3
                fSignal1 = Value
            Case 4
                fSignal2 = Value
        End Select
    End Sub

    Public Overrides Sub AddParameterValue(ByVal Index As Integer, ByVal Value As Double)
        Select Case Index
            Case 0
                fAdd += Value
            Case 1
                fIntensity += Value
            Case 2
                fSwitch += Value
            Case 3
                fSignal1 += Value
            Case 4
                fSignal2 += Value
        End Select
    End Sub

    Public Overrides ReadOnly Property Type As FunctionType
        Get
            Return FunctionType.Function
        End Get
    End Property
End Class
