﻿Public Class Mandelbrot
    Inherits FunctionBase

    Private fAdd As Double = 0
    Private fIntensity As Double = 1

    Private xScale As Double = 2
    Private yScale As Double = 2
    Private xLocation As Double = -0.5
    Private yLocation As Double = 0

    Private Const MAX_ITERATIONS As Integer = 300

    Private ptCenter As New PointF(0.5, 0.5)
    Public Overrides Function Evaluate(ByVal x As Double, ByVal y As Double) As Double

        x = (x - ptCenter.X) * xScale + xLocation
        y = (y - ptCenter.Y) * yScale + yLocation

        Dim tmp As Int32 = CountInterations(x, y)
        Dim fResult As Double = tmp / MAX_ITERATIONS
        Return (fResult + fAdd) * fIntensity
    End Function



    Public Overrides ReadOnly Property Help As String
        Get
            Return String.Format("Generiert die Mandelbrot Menge." &
                "{0}" &
                "Es werden bis zu {1} Iterationen (x * x - y * y) + x0 bzw. (x * y + x * y) + y0 pro Punkt gerechnet solange " &
                "(x * x + y * y) < 4. {0}" &
                "Die Anzahl der Iterationen ergibt dann das Ergebnissignal nach:{0}" &
                "{0}" &
                "Result = (({0}" &
                "   Iterationen / {1} {0}" &
                ") + Add ) * Intensity {0}" &
                "", Environment.NewLine, MAX_ITERATIONS)
        End Get
    End Property

    Public Overrides ReadOnly Property Name As String
        Get
            Return "Mandelbrot"
        End Get
    End Property

    Public Overrides ReadOnly Property ParameterInfos As System.Data.DataTable
        Get
            Dim tbl As New DataTable
            tbl.Columns.Add("Index", GetType(Int32))
            tbl.Columns.Add("Name", GetType(String))
            tbl.Columns.Add("Step", GetType(Double))
            tbl.Columns.Add("Default", GetType(Double))
            tbl.Columns.Add("Current", GetType(Double))
            tbl.Rows.Add(0, "Add", 0.1, 0, fAdd)
            tbl.Rows.Add(1, "Intensity", 0.1, 1, fIntensity)
            tbl.Rows.Add(2, "Scale X", 0.1, 2, xScale)
            tbl.Rows.Add(3, "Scale Y", 0.1, 2, yScale)
            tbl.Rows.Add(4, "Location X", 0.1, -0.5, xLocation)
            tbl.Rows.Add(5, "Location Y", 0.1, 0, yLocation)
            Return tbl
        End Get
    End Property

    Public Overloads Overrides Sub SetParameterValue(ByVal Index As Integer, ByVal Value As Double)
        Select Case Index
            Case 0
                fAdd = Value
            Case 1
                fIntensity = Value
            Case 2
                xScale = Value
            Case 3
                yScale = Value
            Case 4
                xLocation = Value
            Case 5
                yLocation = Value
        End Select
    End Sub

    Public Overrides Sub AddParameterValue(ByVal Index As Integer, ByVal Value As Double)
        Select Case Index
            Case 0
                fAdd += Value
            Case 1
                fIntensity += Value
            Case 2
                xScale += Value
            Case 3
                yScale += Value
            Case 4
                xLocation += Value
            Case 5
                yLocation += Value
        End Select
    End Sub

    Public Overrides ReadOnly Property Type As FunctionType
        Get
            Return FunctionType.Generator
        End Get
    End Property







    Private Function CountInterations(ByVal x As Double, ByVal y As Double) As Int32

        Dim xNow As Double = 0
        Dim yNow As Double = 0

        Dim c As Int32 = 0
        While c < MAX_ITERATIONS + 1 AndAlso Not (xNow * xNow + yNow * yNow > 4)
            c += 1
            Dim yTemp As Double = xNow * yNow * 2
            xNow = xNow * xNow - yNow * yNow + x
            yNow = yTemp + y
        End While

        Return c - 1
    End Function




End Class
