﻿Public Class Sinus
    Inherits FunctionBase

    Private xFrequency As Double = 1
    Private xPhaseShift As Double = 0
    Private yFrequency As Double = 1
    Private yPhaseShift As Double = 0
    Private fAdd As Double = 1
    Private fIntensity As Double = 0.5

    Overrides ReadOnly Property Name() As String
        Get
            Return "Sinus"
        End Get
    End Property

    Public Overrides ReadOnly Property Type As FunctionType
        Get
            Return FunctionType.Generator
        End Get
    End Property


    Public Overrides ReadOnly Property Help As String
        Get
            Return String.Format("Gibt das Produkt eines Horizontal und eines Vertikal verlaufendes Sinus Signal aus.{0}{0}" &
                "Result = (( {0}" &
                "   Sin(x * xScale + xLocation) {0}" &
                "   * {0}" &
                "   Sin(y * yFrequency + yLocation) {0}" &
                ") + Add ) * Intensity", Environment.NewLine)
        End Get
    End Property

    Private Pi2 As Double = Math.PI * 2
    Public Overrides Function Evaluate(ByVal x As Double, ByVal y As Double) As Double
        'neu 1.0.0.1:
        x -= 0.5
        y -= 0.5

        Return (Math.Sin((x * xFrequency + xPhaseShift) * Pi2) *
                Math.Sin((y * yFrequency + yPhaseShift) * Pi2) + fAdd) * fIntensity
    End Function


    Public Overrides Sub SetParameterValue(ByVal Index As Int32, ByVal Value As Double)
        Select Case Index
            Case 0
                fAdd = Value
            Case 1
                fIntensity = Value
            Case 2
                xFrequency = Value
            Case 3
                yFrequency = Value
            Case 4
                xPhaseShift = Value
            Case 5
                yPhaseShift = Value
        End Select
    End Sub


    Public Overrides Sub AddParameterValue(Index As Integer, Value As Double)
        Select Case Index
            Case 0
                fAdd += Value
            Case 1
                fIntensity += Value
            Case 2
                xFrequency += Value
            Case 3
                yFrequency += Value
            Case 4
                xPhaseShift += Value
            Case 5
                yPhaseShift += Value
        End Select
    End Sub

    Overrides ReadOnly Property ParameterInfos() As DataTable
        Get
            Dim tbl As New DataTable
            tbl.Columns.Add("Index", GetType(Int32))
            tbl.Columns.Add("Name", GetType(String))
            tbl.Columns.Add("Step", GetType(Double))
            tbl.Columns.Add("Default", GetType(Double))
            tbl.Columns.Add("Current", GetType(Double))
            tbl.Rows.Add(0, "Add", 0.1, 1, fAdd)
            tbl.Rows.Add(1, "Intensity", 0.1, 0.5, fIntensity)
            tbl.Rows.Add(2, "Scale X", 0.1, 1, xFrequency)
            tbl.Rows.Add(3, "Scale Y", 0.1, 1, yFrequency)
            tbl.Rows.Add(4, "Location X", 0.1, 0, xPhaseShift)
            tbl.Rows.Add(5, "Location Y", 0.1, 0, yPhaseShift)
            Return tbl
        End Get
    End Property
End Class
