﻿Public Class Voronoi
    Inherits FunctionBase


    Private fAdd As Double = 0
    Private fIntensity As Double = 2
    Private cPoints As Int32 = 5
    Private iSeed As Int32 = 1
    Private iType As Int32 = 0

    Private xScale As Double = 1
    Private yScale As Double = 1
    Private xLocation As Double = 0
    Private yLocation As Double = 0

    Private ltAllPoints As List(Of PointF)
    Private dicRows As Dictionary(Of Int32, Dictionary(Of Int32, List(Of PointF)))


    Private Sub AddIndexedPoint(pt As PointF)
        ltAllPoints.Add(pt)
        Dim iRowKey As Int32 = CInt(pt.X * 10)
        Dim iColKey As Int32 = CInt(pt.Y * 10)
        If Not dicRows.ContainsKey(iRowKey) Then
            dicRows.Add(iRowKey, New Dictionary(Of Integer, List(Of PointF)))
        End If
        Dim dicColumns As Dictionary(Of Int32, List(Of PointF)) = dicRows(iRowKey)
        If Not dicColumns.ContainsKey(iColKey) Then
            dicColumns.Add(iColKey, New List(Of PointF))
        End If
        dicColumns(iColKey).Add(pt)
    End Sub

    Private cMisses As Int32
    Public Overrides Function Evaluate(ByVal x As Double, ByVal y As Double) As Double
        If x = 0 AndAlso y = 0 Then
            'initialisieren
            cMisses = 0

            Dim rnd As New Random(iSeed)
            ltAllPoints = New List(Of PointF)
            dicRows = New Dictionary(Of Int32, Dictionary(Of Int32, List(Of PointF)))
            For i As Int32 = 1 To cPoints
                Dim pt As New PointF(CSng(rnd.NextDouble), CSng(rnd.NextDouble))
                For xOffset As Int32 = -1 To 1
                    For yOffset As Int32 = -1 To 1
                        AddIndexedPoint(New PointF(pt.X + xOffset, pt.Y + yOffset))
                    Next
                Next
            Next
        End If

        'alt!!
        'x = (x + xLocation) * xScale
        'y = (y + yLocation) * yScale

        'neu 1.0.0.1
        x = (x - 0.5) * xScale + xLocation
        y = (y - 0.5) * yScale + yLocation

        ''repeat:
        x = x - Math.Floor(x)
        y = y - Math.Floor(y)

        Dim f1 As Double = Double.MaxValue
        Dim f2 As Double = Double.MaxValue
        If cPoints > 15 Then
            'über 15 punkte mit dic suchen ist schneller
            Dim iRowKey As Int32 = CInt(x * 10)
            Dim iColKey As Int32 = CInt(y * 10)
            Dim ltRows As List(Of Dictionary(Of Int32, List(Of PointF))) =
            dicRows.Where(Function(o) o.Key >= iRowKey - 2 AndAlso o.Key <= iRowKey + 2).
            Select(Function(o2) o2.Value).ToList
            Dim ltPointsQuickFiltered As New List(Of PointF)
            For Each row As Dictionary(Of Int32, List(Of PointF)) In ltRows
                Dim ltPoints As List(Of PointF) = row.Where(Function(o) o.Key >= iColKey - 2 AndAlso o.Key <= iColKey + 2).SelectMany(Function(o) o.Value).ToList
                ltPointsQuickFiltered.AddRange(ltPoints)
            Next

            GetDistances(ltPointsQuickFiltered, x, y, f1, f2)

            If f2 > 0.25 Then
                cMisses += 1
                'If cMisses Mod 100 = 0 Then
                '    Debug.Print(cMisses.ToString)
                'End If
                GetDistances(ltAllPoints, x, y, f1, f2)
            End If

        Else
            '15 punkte oder weniger, besser gleich alle rechnen, da sonst zu viele misses
            GetDistances(ltAllPoints, x, y, f1, f2)
        End If



        Dim fDistance As Double
        Select Case iType
            Case 0
                fDistance = f1
            Case 1
                fDistance = 1 - f1
            Case 2
                fDistance = f2
            Case 3
                fDistance = 1 - f2
            Case 4
                fDistance = f2 - f1
            Case 5
                fDistance = f1 - f2
            Case 6
                fDistance = f1 + f2
            Case Else
                iType = 0
        End Select

        Return (fDistance + fAdd) * fIntensity
    End Function

    Private Sub GetDistances(Points As List(Of PointF), x As Double, y As Double, ByRef f1 As Double, ByRef f2 As Double)
        f1 = Double.MaxValue
        f2 = Double.MaxValue
        Dim xDiff As Double
        Dim yDiff As Double
        Dim fDist As Double
        For Each pt As PointF In Points
            xDiff = (x - pt.X)
            yDiff = (y - pt.Y)
            fDist = xDiff * xDiff + yDiff * yDiff
            If fDist < f1 Then
                f2 = f1
                f1 = fDist
            ElseIf fDist < f2 Then
                f2 = fDist
            End If
        Next
        If f2 < Double.MaxValue Then
            f1 = Math.Sqrt(f1)
            f2 = Math.Sqrt(f2)
        End If
    End Sub

    Public Overrides ReadOnly Property Type As FunctionType
        Get
            Return FunctionType.Generator
        End Get
    End Property


    Public Overrides ReadOnly Property Name As String
        Get
            Return "Voronoi"
        End Get
    End Property

    Public Overrides ReadOnly Property Help As String
        Get
            Return String.Format("Gibt ein Voronoi Diagramm aus. Die Anzahl an Punkten kann eingestellt werden, muss aber während " &
                "der Berechnung konstant bleiben. Die Punkte werden zufällig verteilt. {0}" &
                "{0}" &
                "f1 = Distanz zum nächsten Punkt{0}" &
                "f2 = Distanz zum zweitnächsten Punkt{0}" &
                "{0}" &
                "Result = (({0}" &
                "   Typ 0:  f1 {0}" &
                "   Typ 1:  1 - f1 {0}" &
                "   Typ 2:  f2 {0}" &
                "   Typ 3:  1 - f2 {0}" &
                "   Typ 4:  f2 - f1 {0}" &
                "   Typ 5:  f1 - f2 {0}" &
                "   Typ 6:  f1 + f2 {0}" &
                ") + Add ) * Intensity" &
                "{0}" &
                "Das Ergebnis ist in der Grundeinstellung seamless.", Environment.NewLine)
        End Get
    End Property

    Public Overrides ReadOnly Property ParameterInfos As System.Data.DataTable
        Get
            Dim tbl As New DataTable
            tbl.Columns.Add("Index", GetType(Int32))
            tbl.Columns.Add("Name", GetType(String))
            tbl.Columns.Add("Step", GetType(Double))
            tbl.Columns.Add("Default", GetType(Double))
            tbl.Columns.Add("Current", GetType(Double))
            tbl.Rows.Add(0, "Add", 0.1, 0, fAdd)
            tbl.Rows.Add(1, "Intensity", 0.1, 2, fIntensity)
            tbl.Rows.Add(2, "Number of Points", 1, 5, cPoints)
            tbl.Rows.Add(3, "Seed", 1, 1, iSeed)
            tbl.Rows.Add(4, "Type", 1, 0, iType)
            tbl.Rows.Add(5, "Scale X", 0.1, 1, xScale)
            tbl.Rows.Add(6, "Scale Y", 0.1, 1, yScale)
            tbl.Rows.Add(7, "Location X", 0.1, 1, xLocation)
            tbl.Rows.Add(8, "Location Y", 0.1, 1, yLocation)
            Return tbl
        End Get
    End Property



    Public Overrides Sub SetParameterValue(ByVal Index As Integer, ByVal Value As Double)
        Select Case Index
            Case 0
                fAdd = Value
            Case 1
                fIntensity = Value
            Case 2
                cPoints = CInt(Value)
            Case 3
                iSeed = CInt(Value)
            Case 4
                iType = CInt(Value)
            Case 5
                xScale = Value
            Case 6
                yScale = Value
            Case 7
                xLocation = Value
            Case 8
                yLocation = Value
        End Select
    End Sub

    Public Overrides Sub AddParameterValue(Index As Integer, Value As Double)
        Select Case Index
            Case 0
                fAdd += Value
            Case 1
                fIntensity += Value
            Case 2
                cPoints += CInt(Value)
            Case 3
                iSeed += CInt(Value)
            Case 4
                iType += CInt(Value)
            Case 5
                xScale += Value
            Case 6
                yScale += Value
            Case 7
                xLocation += Value
            Case 8
                yLocation += Value
        End Select
    End Sub
End Class
