﻿'ImageSynth
' D.Aue 07.2018

Imports System.Runtime.InteropServices

Class ImageSynth
    Public Functions As New List(Of IFunction)
    Public Repeaters As New List(Of Repeater)
    'Public Players As New List(Of FunctionInstance)

    Event ImageReady(ByVal IsError As Boolean)

    Public Progress As Double
    Public LastMinValue As Double = Double.MaxValue
    Public LastMaxValue As Double = Double.MinValue
    Public IsNormalized As Boolean = False
    Public Histogram() As Double

    Private dwData() As Int32
    Public OutputImage As Bitmap
    Private cpWidth As Int32
    Private cpHeight As Int32
    Private hData As GCHandle

    Private oCalcThread As Threading.Thread
    Private _isRunning As Boolean
    Private isResetFlag As Boolean

    Private _t As Double

    Public Sub GenerateImageAsync(ByVal Width As Int32, ByVal Height As Int32, Optional ByVal t As Double = 0)
        AbortCalculation()
        Prepare(Width, Height)
        isResetFlag = False
        _isRunning = True
        _t = t
        oCalcThread = New Threading.Thread(AddressOf CalcImage)
        oCalcThread.Start()
    End Sub


    Private ltSequenceData As New List(Of Int32())
    Public Sub GenerateSequence(ByVal Width As Int32, ByVal Height As Int32)
        ltSequenceData.Clear()
        For i As Int32 = 0 To 20
            GenerateImageAsync(Width, Height, i / 20)
            Do While _isRunning
                Threading.Thread.Sleep(10)
            Loop
            ltSequenceData.Add(dwData.ToArray)
        Next
    End Sub
    Public Sub ShowSequenceImage(ByVal Index As Int32)
        If Index >= 0 AndAlso Index < ltSequenceData.Count - 1 Then
            Array.Copy(ltSequenceData(Index), dwData, dwData.Length)
            RaiseEvent ImageReady(False)
        End If
    End Sub

    Public Sub AbortCalculation()
        If _isRunning Then
            isResetFlag = True
            Do While _isRunning
                Threading.Thread.Sleep(10)
            Loop
        End If
    End Sub

    ReadOnly Property IsRunning As Boolean
        Get
            Return _isRunning
        End Get
    End Property

    Private Sub CalcImage()
        Dim isError As Boolean = False
        Dim oSoloInstance As IFunction = Functions.FirstOrDefault(Function(o) o.IsSolo)
        'alle muted funktionen, wo deren output hingeht, den parameter dieser funktion auf Default setzen
        Dim ltMuted As List(Of IFunction) = Functions.Where(Function(o) o.IsMute).ToList
        ltMuted.ForEach(Sub(o)
                            o.Outputs.ForEach(Sub(o2)
                                                  Dim tbl As DataTable = o2.ToFunction.ParameterInfos
                                                  o2.ToFunction.SetParameterValue(o2.ToPin, CDbl(tbl.Rows(o2.ToPin)("Default")))
                                              End Sub)
                        End Sub)


        SyncLock Functions
            LastMinValue = Double.MaxValue
            LastMaxValue = Double.MinValue
            Array.Clear(Histogram, 0, Histogram.Length)

            Dim fResults(cpWidth * cpHeight - 1) As Double


            For y As Int32 = 0 To cpHeight - 1
                Progress = y / cpHeight
                For x As Int32 = 0 To cpWidth - 1

                    Functions.ForEach(Sub(o) o.ProvidedInputCount = 0)

                    'Alle wo ein input signal dranhängt auf 0 setzen
                    Functions.Where(Function(o) Not o.IsMute AndAlso o.Type <> FunctionType.Player).
                              SelectMany(Function(o) o.Outputs).ToList.
                              ForEach(Sub(o) o.ZeroValue())


                    Dim isFinished As Boolean = False
                    Do While Functions.Any(Function(o) o.ProvidedInputCount <= o.RequiredInputCount)
                        Dim ltProcessable As List(Of IFunction) = Functions.Where(
                            Function(o)
                                Return (o.ProvidedInputCount = o.RequiredInputCount AndAlso
                                        (
                                            (o.Outputs.Count > 0 OrElse (o.Type = FunctionType.ImageOutput AndAlso oSoloInstance Is Nothing)) OrElse
                                            (o.Equals(oSoloInstance))
                                        )
                                       )
                            End Function).ToList
                        If ltProcessable.Count = 0 Then
                            isError = True
                            isResetFlag = True
                            Exit Do
                        End If
                        For Each oFunc As IFunction In ltProcessable
                            If oFunc.IsMute Then
                                For Each oChild As OutputTo In oFunc.Outputs
                                    oChild.ToFunction.ProvidedInputCount += 1
                                Next
                            Else
                                Dim oTemp As IFunction = oFunc
                                Dim ltConnectedRepeaters As List(Of Repeater) = Repeaters.Where(Function(o) o.Outputs.Any(Function(o2)
                                                                                                                              Return o2.ToFunction.Equals(oTemp)
                                                                                                                          End Function)).ToList
                                'Ergebniswert berechnen, ggf. berechnung x-fach wiederholen
                                Dim fThisFuncResult As Double = 0
                                If ltConnectedRepeaters.Count = 0 Then
                                    fThisFuncResult = oFunc.Evaluate(x / cpWidth, y / cpHeight)
                                Else
                                    ltConnectedRepeaters.SelectMany(Function(o) o.Outputs.Where(Function(o2) o2.ToFunction.Equals(oTemp))).ToList.
                                        ForEach(Sub(o2)
                                                    o2.ZeroValue()
                                                End Sub)
                                    Dim iIteration As Int32 = 1
                                    Dim isStillIterating As Boolean = True
                                    Do While isStillIterating
                                        isStillIterating = False
                                        For Each oRep As Repeater In ltConnectedRepeaters
                                            Dim tbl As DataTable = oRep.DataSequence
                                            If tbl.Rows.Count >= iIteration Then
                                                Dim row As DataRow = tbl.Rows(iIteration - 1)
                                                For Each oOutput As OutputTo In oRep.Outputs.Where(Function(o) o.ToFunction.Equals(oTemp)).ToList
                                                    Dim oValue As Object = row(oOutput.FromPin)
                                                    If Not oValue.Equals(DBNull.Value) Then
                                                        oFunc.SetParameterValue(oOutput.ToPin, CDbl(oValue))
                                                    End If
                                                Next
                                                isStillIterating = True 'weitermachen solange es repeater mit genügend rows gibt
                                            End If
                                        Next
                                        If iIteration = 1 OrElse isStillIterating Then
                                            fThisFuncResult += oFunc.Evaluate(x / cpWidth, y / cpHeight)
                                            iIteration += 1
                                        End If
                                    Loop
                                End If

                                'Ergebnis an folge funktionen weitergeben
                                For Each oChild As OutputTo In oFunc.Outputs
                                    oChild.ToFunction.AddParameterValue(oChild.ToPin, fThisFuncResult)
                                    oChild.ToFunction.ProvidedInputCount += 1
                                Next
                                isFinished = (oFunc.Type = FunctionType.ImageOutput AndAlso oSoloInstance Is Nothing) OrElse oFunc.Equals(oSoloInstance)

                                If isFinished Then
                                    fResults(y * cpWidth + x) = fThisFuncResult
                                    If LastMaxValue < fThisFuncResult Then LastMaxValue = fThisFuncResult
                                    If LastMinValue > fThisFuncResult Then LastMinValue = fThisFuncResult
                                End If
                            End If
                            oFunc.ProvidedInputCount = Int32.MaxValue
                            If isFinished Then Exit For
                        Next
                        If isFinished Then Exit Do 'Solo oder Image gerendert->aus
                    Loop
                    If isResetFlag Then Exit For
                Next
                If isResetFlag Then Exit For

                'diese Zeile ins bmp übertragen
                Dim bt As Int32
                For xPixel As Int32 = 0 To cpWidth - 1
                    Dim fValue As Double = fResults(y * cpWidth + xPixel)
                    If oSoloInstance Is Nothing Then
                        dwData(y * cpWidth + xPixel) = CInt(fValue)
                    Else
                        If fValue > 1 Then
                            fValue = 1
                        ElseIf fValue < 0 Then
                            fValue = 0
                        End If
                        bt = CInt(fValue * 255) And &HFF
                        dwData(y * cpWidth + xPixel) = bt << 16 Or bt << 8 Or bt
                    End If
                Next

            Next

            If Not isResetFlag Then
                Dim fStep As Double = (LastMaxValue - LastMinValue) / (Histogram.Count - 1)
                Dim fPercent As Double = 1 / (cpWidth * cpHeight)
                If fStep > 0 Then
                    Array.ForEach(fResults, Sub(f) Histogram(CInt((f - LastMinValue) / fStep)) += fPercent)
                End If
            End If

            IsNormalized = False
            If Not isResetFlag AndAlso oSoloInstance IsNot Nothing Then
                'solo -> normalisieren?
                If LastMaxValue = Double.MinValue Then LastMaxValue = 0
                If LastMinValue = Double.MaxValue Then LastMinValue = 0

                IsNormalized = ((LastMinValue > 0.7) OrElse (LastMaxValue < 0.3)) AndAlso Not (LastMaxValue = LastMinValue)

                Dim fNormalize As Double = 1 / (LastMaxValue - LastMinValue)

                If IsNormalized Then
                    Dim bt As Int32
                    Dim fResult As Double
                    For yPixel As Int32 = 0 To cpHeight - 1
                        For xPixel As Int32 = 0 To cpWidth - 1
                            fResult = fResults(yPixel * cpWidth + xPixel)
                            fResult = (fResult - LastMinValue) * fNormalize
                            If fResult > 1 Then
                                fResult = 1
                            ElseIf fResult < 0 Then
                                fResult = 0
                            End If
                            bt = CInt(fResult * 255) And &HFF
                            dwData(yPixel * cpWidth + xPixel) = bt << 16 Or bt << 8 Or bt
                        Next
                    Next
                End If
            End If

            _isRunning = False
        End SyncLock
        Progress = 1
        RaiseEvent ImageReady(isError)
    End Sub




    Private Sub Prepare(ByVal Width As Int32, ByVal Height As Int32)
        If OutputImage IsNot Nothing AndAlso (Width <> cpWidth OrElse Height <> cpHeight) Then
            CleanUp()
        End If

        If OutputImage Is Nothing Then
            cpWidth = Width
            cpHeight = Height
            Dim cbStride As Int32 = (cpWidth * 4 + 3) And &HFFFFFFFC

            ReDim dwData(cpWidth * cpHeight - 1)
            hData = GCHandle.Alloc(dwData, GCHandleType.Pinned)
            OutputImage = New Bitmap(cpWidth, cpHeight, cbStride, Imaging.PixelFormat.Format32bppRgb, hData.AddrOfPinnedObject)

            ReDim Histogram(cpWidth - 1)
        End If
    End Sub

    Public Sub CleanUp()
        If OutputImage IsNot Nothing Then
            OutputImage.Dispose()
            OutputImage = Nothing
            hData.Free()
        End If
    End Sub
End Class
