﻿'ImageSynth
' D.Aue 07.2018
' V 1.0.0.1 Location/Scale bei allen Generatoren angepasst, Histogramm ergänzt.

Public Class frmMain
    Private WithEvents oSynth As New ImageSynth

    Private ptMouse As Point
    Private ptMouseDown As Point

    Private sLoadedFile As String

    Private Sub Form1_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        ConfigManager.Client = pnlConfig
        SetupTest()

        UpdateWindowText()

        UpdateImage()
        DrawConfig()
    End Sub


    Private Sub UpdateWindowText()
        Dim sFile As String = IO.Path.GetFileName(sLoadedFile)
        Me.Text = "Image Synth Version " & Application.ProductVersion & If(sFile <> "", " - " & sFile, "")
    End Sub


    Private Sub AddFunctionButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnAddFunction.Click
        Dim frm As New frmChooseFunction
        frm.ShowDialog()
        If frm.SelectedFunction IsNot Nothing Then
            Dim oNewItem As New ConfigInstance With {.Function = frm.SelectedFunction}
            AddModule(oNewItem)
            DrawConfig()
        End If
    End Sub

    Private Sub SetupTest()
        Dim oNoise As New SimplexNoise
        oNoise.SetParameterValue("Intensity", 1)
        Dim oSinus As New Sinus
        Dim oImage As New ImageOutput

        ' Noise schickt output an sinus
        oNoise.AddOutput(New OutputTo With {.ToFunction = oSinus, .ToPin = 0})

        ' sinus schickt an image RGB
        oSinus.AddOutput(New OutputTo With {.ToFunction = oImage, .ToPin = 0})
        oSinus.AddOutput(New OutputTo With {.ToFunction = oImage, .ToPin = 1})
        oSinus.AddOutput(New OutputTo With {.ToFunction = oImage, .ToPin = 2})

        oSynth.Functions.AddRange({oNoise, oSinus, oImage})

        'oSynth.Players.Add(New FunctionInstance With {.Function = New LinearPlayer})

        AddInstance(New ConfigInstance With {.Function = oNoise})
        AddInstance(New ConfigInstance With {.Function = oSinus})
        AddInstance(New ConfigInstance With {.Function = oImage})

        Dim pt As New Point(10, 10)
        ConfigInstances.ForEach(Sub(o)
                                    o.Rect.Location = pt
                                    pt.Offset(250, 50)
                                End Sub)
    End Sub

    Private Sub UpdateImage()
        oSynth.GenerateImageAsync(picOutput.Width, picOutput.Height)
    End Sub


    Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        oSynth.CleanUp()
    End Sub

    Private Sub picConfig_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles picConfig.Paint
        PaintConfig(e.Graphics, ptMouse)
    End Sub

    Private Sub btnHelp_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnHelp.Click
        frmHelp.Show()
    End Sub




#Region "Mousehandling über Konfiguration"
    Private Sub picConfig_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picConfig.MouseDoubleClick
        If ConfigInstances Is Nothing Then Exit Sub

        Dim oInstance As ConfigInstance = ConfigInstances.LastOrDefault(Function(o) o.IsHit(e.Location))
        If oInstance IsNot Nothing Then
            Dim iHittestResult As ConfigInstance.HittestResult = oInstance.Hittest(e.Location)
            Select Case iHittestResult And &HFF00
                Case ConfigInstance.HittestResult.ParameterValue, ConfigInstance.HittestResult.ParameterName
                    'doppelklick auf parametername/value
                    If oInstance.Function.Type = FunctionType.Repeater Then
                        Dim oRep As Repeater = CType(oInstance.[Function], Repeater)
                        Dim frm As New frmRepeaterValues
                        frm.Display(oRep, AddressOf UpdateImage)
                    Else
                        Dim iParameter As Int32 = iHittestResult And &HFF
                        Dim frm As New frmEnterValue
                        frm.Display(oInstance, iParameter)

                        frm.ShowDialog()
                        If frm.IsValid Then
                            oInstance.[Function].SetParameterValue(iParameter, frm.Value)
                        End If
                    End If

                    UpdateImage()
                    DrawConfig()
            End Select
        End If
    End Sub


    Private Sub picConfig_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picConfig.MouseDown
        If ConfigInstances Is Nothing Then Exit Sub

        ptMouseDown = e.Location

        Dim oInstance As ConfigInstance = ConfigInstances.LastOrDefault(Function(o) o.IsHit(ptMouseDown))
        If oInstance IsNot Nothing Then
            Dim iHittestResult As ConfigInstance.HittestResult = oInstance.Hittest(ptMouseDown)
            Dim isHandled As Boolean = False
            Dim isRedraw As Boolean = False
            Select Case iHittestResult
                Case ConfigInstance.HittestResult.Delete
                    'Löschen
                    RemoveModule(oInstance)
                    isRedraw = True

                Case ConfigInstance.HittestResult.Info
                    Dim frm As New frmModuleInfo
                    frm.lblText.Text = oInstance.[Function].Help
                    frm.Show()
                    isHandled = True

                Case ConfigInstance.HittestResult.Solo
                    ToggleSolo(oInstance)
                    isRedraw = True

                Case ConfigInstance.HittestResult.Mute
                    ToggleMute(oInstance)
                    isRedraw = True

                Case ConfigInstance.HittestResult.OutputConnector
                    'neue verbindung am output
                    oInstance.IsNewOutputAdd = True
                    oInstance.NewOutputAddPin = -1
                    isHandled = True

                Case Is > ConfigInstance.HittestResult.OutputConnector
                    Dim iParameter As Int32 = iHittestResult And &HFF
                    Select Case iHittestResult And &HFF00
                        'Parameter zeugs
                        Case ConfigInstance.HittestResult.ParameterInc, ConfigInstance.HittestResult.ParameterDec
                            '+/-
                            HandleParameterMouseDown(oInstance, iParameter, CType((iHittestResult And &HFF00), ConfigInstance.HittestResult))
                            isRedraw = True

                        Case ConfigInstance.HittestResult.InputConnector
                            If oInstance.Function.Type = FunctionType.Repeater Then
                                'neue verbindung am Ausgang eines Repeaters
                                oInstance.IsNewOutputAdd = True
                                oInstance.NewOutputAddPin = iParameter
                                isHandled = True
                            Else
                                If oInstance.Function.Name = "Add" Then
                                    iParameter = 1  'beim add immer den Signal Parameter nehmen
                                End If

                                'klick an parameter connector -> ist da was verbunden?
                                Dim oInstanceConnectedFrom As ConfigInstance
                                oInstanceConnectedFrom = ConfigInstances.Where(
                                        Function(o) o.[Function].Outputs.Any(
                                            Function(o2) o2.ToFunction.Equals(oInstance.[Function]) _
                                                         AndAlso o2.ToPin = iParameter
                                        )).OrderBy(Function(o) o.ConnectorHitscore(ptMouseDown, oInstance)).FirstOrDefault



                                If oInstanceConnectedFrom IsNot Nothing Then
                                    'mit diesem input connector ist etwas verbunden -> Verbindung lösen und Drag Start
                                    Dim oOutput As OutputTo = oInstanceConnectedFrom.[Function].Outputs.FirstOrDefault(
                                            Function(o) o.ToFunction.Equals(oInstance.[Function]) AndAlso
                                                        o.ToPin = iParameter)
                                    HandleDisconnect(oInstanceConnectedFrom, oOutput)
                                    isRedraw = True
                                Else
                                    'mit diesem input connector ist nichts verbunden -> ignorieren
                                End If
                            End If
                    End Select
                Case Else
            End Select

            isHandled = isHandled Or isRedraw
            If Not isHandled Then
                If e.Button = MouseButtons.Left Then
                    'Drag start
                    oInstance.IsDrag = True
                    oInstance.MouseDown = New Point(oInstance.Rect.X - e.Location.X, oInstance.Rect.Y - e.Location.Y)
                    DrawConfig()
                ElseIf e.Button = MouseButtons.Right Then
                    'rechtsklick -> Modul duplizieren und drag start
                    If oInstance.Function.Type = FunctionType.Function OrElse oInstance.Function.Type = FunctionType.Generator Then
                        Dim oNew As ConfigInstance = oInstance.Clone
                        oNew.Function.IsSolo = False
                        oNew.IsDrag = True
                        oNew.MouseDown = New Point(oInstance.Rect.X - e.Location.X, oInstance.Rect.Y - e.Location.Y)
                        AddModule(oNew)
                    End If
                End If
            ElseIf isRedraw Then
                UpdateImage()
                DrawConfig()
            End If
        End If
    End Sub



    Private Sub picConfig_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picConfig.MouseMove
        If ConfigInstances Is Nothing Then Exit Sub

        ptMouse = e.Location

        Dim oInstance As ConfigInstance

        oInstance = ConfigInstances.FirstOrDefault(Function(o) o.IsDrag)
        If oInstance IsNot Nothing Then
            Dim ptNew As Point = ClampLocation(ptMouse)
            oInstance.Rect.Location = New Point(ptNew.X + oInstance.MouseDown.X, ptNew.Y + oInstance.MouseDown.Y)
            DrawConfig()
        End If

        oInstance = ConfigInstances.FirstOrDefault(Function(o) o.IsNewOutputAdd)
        If oInstance IsNot Nothing Then
            'drag eines neuen output connectors
            picConfig.Invalidate()
        End If

        oInstance = ConfigInstances.FirstOrDefault(Function(o) o.IsParameterDrag)
        If oInstance IsNot Nothing AndAlso e.Button = MouseButtons.Left Then
            'drag einer Parameter value
            Dim yDiff As Int32 = oInstance.MouseDown.Y - ptMouse.Y
            Dim iSteps As Int32 = CInt(yDiff / 10)

            If iSteps <> oInstance.PrevParameterStep Then
                ScrollParameter(oInstance.Function, oInstance.ParameterIndexAtMousedown, -(oInstance.PrevParameterStep - iSteps))
                oInstance.PrevParameterStep = iSteps
                UpdateImage()
                DrawConfig()
            End If
        End If

    End Sub

    Private Sub picConfig_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles picConfig.MouseUp
        If ConfigInstances Is Nothing Then Exit Sub

        Dim isRedrawImage As Boolean

        Dim oInstance As ConfigInstance

        oInstance = ConfigInstances.FirstOrDefault(Function(o) o.IsDrag)
        If oInstance IsNot Nothing Then
            Dim ptNew As Point = ClampLocation(ptMouse)
            oInstance.Rect.Location = New Point(ptNew.X + oInstance.MouseDown.X, ptNew.Y + oInstance.MouseDown.Y)
            oInstance.IsDrag = False
            DrawConfig()
        End If

        oInstance = ConfigInstances.FirstOrDefault(Function(o) o.IsNewOutputAdd)
        If oInstance IsNot Nothing Then
            'drop eines neuen output connectors
            Dim oInstanceConnectedTo As ConfigInstance = ConfigInstances.LastOrDefault(Function(o) o.IsHit(e.Location))
            If oInstanceConnectedTo IsNot Nothing AndAlso Not oInstance.Equals(oInstanceConnectedTo) Then
                Dim iHittestResult As ConfigInstance.HittestResult = oInstanceConnectedTo.Hittest(e.Location)
                If (iHittestResult And &HFF00) = ConfigInstance.HittestResult.InputConnector Then
                    'Mouse up auf einem Input Connector
                    Dim iParameter As Int32 = iHittestResult And &HFF
                    'connection hinzufügen

                    HandleConnect(oInstance, oInstance.NewOutputAddPin, oInstanceConnectedTo, iParameter)
                    isRedrawImage = True
                End If
            End If

            oInstance.IsNewOutputAdd = False
            If isRedrawImage Then
                UpdateImage()
            End If
            DrawConfig()
        End If


        oInstance = ConfigInstances.FirstOrDefault(Function(o) o.IsParameterDrag)
        If oInstance IsNot Nothing Then
            oInstance.IsParameterDrag = False
        End If

    End Sub


    Private Sub PicMouseWheel(ByVal sender As Object, ByVal e As MouseEventArgs) Handles picConfig.MouseWheel
        Dim oInstance As ConfigInstance = ConfigInstances.LastOrDefault(Function(o) o.IsHit(e.Location))
        If oInstance IsNot Nothing Then
            Dim iHittestResult As ConfigInstance.HittestResult = oInstance.Hittest(e.Location)
            Select Case iHittestResult And &HFF00
                Case ConfigInstance.HittestResult.ParameterInc, ConfigInstance.HittestResult.ParameterDec

                    Dim iParameter As Int32 = iHittestResult And &HFF
                    '+/-
                    ScrollParameter(oInstance.Function, iParameter, CInt(e.Delta / 120))

                    UpdateImage()
                    DrawConfig()
            End Select
        End If
    End Sub

    Private Function ClampLocation(ByVal MouseLocation As Point) As Point
        Dim ptNew As Point
        ptNew.X = If(MouseLocation.X < 0, 0, MouseLocation.X)
        ptNew.Y = If(MouseLocation.Y < 0, 0, MouseLocation.Y)
        Return ptNew
    End Function

    Private Sub pnlConfig_SizeChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles pnlConfig.SizeChanged
        If Me.IsHandleCreated And Not Me.WindowState = FormWindowState.Minimized Then
            DrawConfig()
        End If
    End Sub
#End Region



    Private Sub btnSize_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnSize.Click
        If btnSize.Text = ">" Then
            btnSize.Text = "<"
            pnlOutput.Width = Me.ClientSize.Width - picOutput.Left - 20
            pnlOutput.Height = Me.ClientSize.Height - picOutput.Top - btnSize.Height
        Else
            btnSize.Text = ">"
            pnlOutput.Location = New Point(12, 12)
            pnlOutput.Size = New Size(279, 306)
        End If
        picOutput.Image = Nothing
        For Each ctl As Control In {CType(pnlConfig, Control), btnAddFunction, btnLoad, btnSave, btnHelp, picProgress, lblProgress}
            ctl.Visible = Not ctl.Visible
        Next
        UpdateImage()
        picOutput.Image = oSynth.OutputImage
    End Sub



    Private Sub oSynth_ImageReady(ByVal IsError As Boolean) Handles oSynth.ImageReady
        If picOutput.InvokeRequired Then
            Me.Invoke(Sub() oSynth_ImageReady(IsError))
        Else
            If IsError Then
                MessageBox.Show("Fehler bei der Berechnung!")
            Else
                If Me.IsHandleCreated Then
                    picOutput.Image = oSynth.OutputImage
                End If
            End If
        End If
    End Sub


    Private Sub picOutput_Paint(ByVal sender As Object, ByVal e As PaintEventArgs) Handles picOutput.Paint
        Dim oInstance As ConfigInstance = ConfigInstances.FirstOrDefault(Function(o) o.[Function].IsSolo)
        If oInstance IsNot Nothing Then
            'ein Solo ist aktiv -> min und max werte einblenden
            Dim sText As String = String.Format("Min: {0:N3}, Max: {1:N3}", oSynth.LastMinValue, oSynth.LastMaxValue)
            Dim oFont As Font = New Font("Microsoft Sans Serif", 15, GraphicsUnit.Pixel)
            Dim fSize As SizeF = e.Graphics.MeasureString(sText, oFont)
            Dim rect As Rectangle = New Rectangle(10, 10, CInt(fSize.Width), CInt(fSize.Height))
            e.Graphics.FillRectangle(New SolidBrush(Color.FromArgb(150, 0, 0, 0)), rect)
            e.Graphics.DrawString(sText, oFont, Brushes.HotPink, New Point(10, 10))
            If oSynth.IsNormalized Then
                sText = "Normalized"
                fSize = e.Graphics.MeasureString(sText, oFont)
                rect = New Rectangle(10, rect.Bottom + 2, CInt(fSize.Width), CInt(fSize.Height))
                e.Graphics.FillRectangle(New SolidBrush(Color.FromArgb(150, 0, 0, 0)), rect)
                e.Graphics.DrawString(sText, oFont, Brushes.LightGreen, New Point(10, rect.Top))
            End If

            'Histogramm
            Dim xPos As Int32 = 0
            e.Graphics.DrawLines(Pens.Red, oSynth.Histogram.Select(Function(f)
                                                                       Dim pt As New Point(xPos, CInt(picOutput.Height - (f * 10 * picOutput.Height)))
                                                                       xPos += 1
                                                                       Return pt
                                                                   End Function).ToArray)
        End If
    End Sub

    Dim dtCalcfinished As DateTime
    Dim cSequenceImage As Int32
    Private Sub tmr_Tick(ByVal sender As Object, ByVal e As EventArgs) Handles tmr.Tick
        If oSynth.IsRunning Then
            lblProgress.Visible = True
            picProgress.Visible = True
            dtCalcfinished = DateTime.MinValue
            picProgress.Invalidate()
            picOutput.Invalidate()
        ElseIf dtCalcfinished = DateTime.MinValue Then
            picProgress.Invalidate()
            dtCalcfinished = DateTime.Now
        ElseIf (DateTime.Now - dtCalcfinished).TotalSeconds > 1 Then
            lblProgress.Visible = False
            picProgress.Visible = False
        End If
        If btnAnimate.Text = "Stop" Then
            oSynth.ShowSequenceImage(cSequenceImage)
            cSequenceImage += 1
            If cSequenceImage >= 20 Then cSequenceImage = 0
        End If
    End Sub


    Private Sub picProgress_Paint(ByVal sender As Object, ByVal e As PaintEventArgs) Handles picProgress.Paint
        e.Graphics.Clear(Color.Empty)
        e.Graphics.FillRectangle(Brushes.LightGreen, New Rectangle(0, 0, CInt(picProgress.Width * oSynth.Progress), picProgress.Height))
    End Sub








    Private Sub btnSave_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnSave.Click
        oSynth.AbortCalculation()

        Dim xml As New Xml.XmlDocument
        Dim ndMain As Xml.XmlNode = xml.CreateElement("Main")
        xml.AppendChild(ndMain)

        ConfigInstances.SelectMany(Function(o) o.[Function].Outputs).ToList.
                ForEach(Sub(o)
                            o.PrepareSave()
                        End Sub)

        For Each oInstance As ConfigInstance In ConfigInstances
            oInstance.Serialize(ndMain)
        Next

        Dim dlg As New SaveFileDialog
        dlg.Filter = ".xml|*.xml"
        If sLoadedFile <> "" Then
            dlg.FileName = IO.Path.GetFileName(sLoadedFile)
        End If
        dlg.ShowDialog()

        Dim sPath As String = dlg.FileName

        If Not sPath.ToLower.EndsWith(".xml") Then
            sPath &= ".xml"
        End If

        xml.Save(sPath)
        sLoadedFile = sPath
        UpdateWindowText()
    End Sub

    Private Sub btnLoad_Click(ByVal sender As Object, ByVal e As EventArgs) Handles btnLoad.Click
        oSynth.AbortCalculation()

        Dim dlg As New OpenFileDialog
        dlg.Filter = ".xml|*.xml"
        dlg.ShowDialog()

        sLoadedFile = dlg.FileName

        If sLoadedFile.ToLower.EndsWith(".xml") Then
            Dim xml As New Xml.XmlDocument
            xml.Load(sLoadedFile)
            ConfigInstances.Clear()


            oSynth.AbortCalculation()
            SyncLock oSynth.Functions
                oSynth.Functions.Clear()
            End SyncLock

            For Each nd As Xml.XmlElement In xml.SelectNodes("Main/ConfigInstance")
                Dim oInstance As New ConfigInstance(nd)
                AddModule(oInstance)
                'AddInstance(oInstance)
                'ConfigInstances.Add(oInstance)
                'If oInstance.Function.Type = FunctionType.Repeater Then
                '    oSynth.Repeaters.Add(CType(oInstance.[Function], Repeater))
                'Else
                '    oSynth.Functions.Add(oInstance.[Function])
                'End If
            Next


            'jetzt sind alle instanzen verfügbar->die connections aufbauen
            ConfigInstances.ForEach(Sub(o)
                                        o.Function.Outputs.ForEach(Sub(o2)
                                                                       If Not o2.InitAfterLoad(ConfigInstances, o.Function) Then
                                                                           o.[Function].Outputs.Remove(o2)
                                                                       End If
                                                                   End Sub)
                                    End Sub)
            UpdateImage()
            DrawConfig()
            UpdateWindowText()
        End If

    End Sub













    Private Sub AddModule(Instance As ConfigInstance)
        AddFunction(Instance.Function)

        If Instance.Rect.IsEmpty Then
            Dim pt As New Point(pnlConfig.HorizontalScroll.Value + 10, pnlConfig.VerticalScroll.Value + 10)
            Instance.Rect.Location = pt
            Instance.UpdateParametersAndRectSize()
            Do While ConfigInstances.Any(Function(o) o.Rect.Location = pt)
                pt.Offset(10, 10)
            Loop
            Instance.Rect.Location = pt
        End If

        AddInstance(Instance)
    End Sub

    Private Sub AddFunction(ByVal Func As IFunction)
        oSynth.AbortCalculation()
        SyncLock oSynth.Functions
            If {FunctionType.Generator, FunctionType.Function, FunctionType.ImageOutput}.Contains(Func.Type) Then
                oSynth.Functions.Add(Func)
            ElseIf Func.Type = FunctionType.Repeater Then
                oSynth.Repeaters.Add(CType(Func, Repeater))
            End If
        End SyncLock
    End Sub

    Private Sub RemoveModule(ByVal Instance As ConfigInstance)
        oSynth.AbortCalculation()
        SyncLock oSynth.Functions
            ConfigInstances.Remove(Instance)
            oSynth.Functions.Remove(Instance.[Function])

            'alles was am output hängt lösen
            Instance.[Function].Outputs.ToList.ForEach(Sub(o) Instance.[Function].RemoveOutput(o))

            'alles was am input hängt lösen
            Dim ltReceivingFrom As List(Of IFunction) = oSynth.Functions.Where(Function(o) o.Outputs.Any(Function(o2)
                                                                                                             Return o2.ToFunction.Equals(Instance.[Function])
                                                                                                         End Function)).ToList
            ltReceivingFrom.ForEach(Sub(o)
                                        Dim ltOutputs As List(Of OutputTo) = o.Outputs.Where(Function(o2) o2.ToFunction.Equals(Instance.[Function])).ToList
                                        ltOutputs.ForEach(Sub(o2) o.RemoveOutput(o2))
                                    End Sub)
        End SyncLock
    End Sub


    Private Sub HandleParameterMouseDown(ByVal Instance As ConfigInstance, ByVal ParameterIndex As Int32, ByVal Direction As ConfigInstance.HittestResult)
        ScrollParameter(Instance.Function, ParameterIndex, If(Direction = ConfigInstance.HittestResult.ParameterInc, 1, -1))

        Instance.IsParameterDrag = True
        Instance.MouseDown = ptMouseDown
        Instance.PrevParameterStep = 0
        Instance.ParameterIndexAtMousedown = ParameterIndex
    End Sub

    Private Sub HandleDisconnect(ByVal FromInstance As ConfigInstance, ByVal Output As OutputTo)

        oSynth.AbortCalculation()
        SyncLock oSynth.Functions
            FromInstance.[Function].RemoveOutput(Output)
        End SyncLock

        FromInstance.IsNewOutputAdd = True
        FromInstance.NewOutputAddPin = Output.FromPin
    End Sub

    Private Sub HandleConnect(ByVal FromInstance As ConfigInstance, FromPin As Int32, ToInstance As ConfigInstance, ToPin As Int32)
        If ToInstance.Function.Name = "Add" Then
            ToPin = 1   'add immer am Value2 verbinden
        End If

        Dim oNewOutput As New OutputTo With {.ToFunction = ToInstance.Function,
                                                                .ToPin = ToPin,
                                                                .FromPin = FromPin}
        oSynth.AbortCalculation()
        SyncLock oSynth.Functions
            FromInstance.[Function].AddOutput(oNewOutput)
        End SyncLock
    End Sub


    Private Sub ToggleSolo(ByVal Instance As ConfigInstance)
        Dim oPrevSolo As ConfigInstance = ConfigInstances.FirstOrDefault(Function(o) o.[Function].IsSolo = True AndAlso Not o.Equals(Instance))
        If oPrevSolo IsNot Nothing Then
            oPrevSolo.[Function].IsSolo = False
        End If
        Instance.[Function].IsSolo = Not Instance.[Function].IsSolo
        If Instance.[Function].IsSolo AndAlso Instance.[Function].IsMute Then
            Instance.[Function].IsMute = False
        End If
    End Sub

    Private Sub ToggleMute(ByVal Instance As ConfigInstance)
        Instance.[Function].IsMute = Not Instance.[Function].IsMute
        If Instance.[Function].IsSolo AndAlso Instance.[Function].IsMute Then
            Instance.[Function].IsSolo = False
        End If
    End Sub

    Private Sub ScrollParameter(ByVal Func As IFunction, ByVal ParameterIndex As Int32, ByVal Steps As Int32)
        Dim tbl As DataTable = Func.ParameterInfos
        Dim fValue As Double = CDbl(tbl.Rows(ParameterIndex)("Current"))
        Dim fStep As Double = CDbl(tbl.Rows(ParameterIndex)("Step"))

        If fStep = 1 Then
            Func.SetParameterValue(ParameterIndex, fValue + fStep * Steps)
        Else
            fStep = Math.Abs(fValue * fStep) * Steps
            If Math.Abs(fStep) < 0.001 Then
                fStep = 0.001 * Math.Sign(Steps)
            End If
            Func.SetParameterValue(ParameterIndex, fValue + fStep)
        End If


    End Sub
End Class
